import os
from time import perf_counter
from biochemHH import GBparser
from biochemHH.GBparser import reverse_complement
from biochemHH.GBparser import preferred_codon
from biochemHH.GBparser import preferred_codon_split

if __name__ == '__main__':
    wd = 'C:/Users/hsinmei/playground';
    os.chdir(wd)

    """
    This example file consist of the following parts:
    1. Print notes
    2. Cloning design (advanced)
        - Example 0a: mutate BsaI site on a vector
        - Example 0b: subclone a gene onto a vector
        - Example 0c: construct a fusion protein involving primer walk 
        - Example 0d: subclone a gene onto a vector & mutate BsaI site simultaneously
    """

    # ================================================================================ 1. Print notes
    # { ........................................ Print notes (advanced)
    GBparser.print_note('Cloning design (advanced)')
    GBparser.print_note('About oligo & plasmid list')
    GBparser.print_note('About golden gate')
    GBparser.print_note('Primer design explained')
    # }

    # ================================================================================ 2. Cloning design (advanced)

    # ................................................................................ Example 0a: mutate BsaI site on a vector
    """
    1. Edit the sequence (create a new gb file with the BsaI site replaced) -> 'pUCv2'
    2. Cloning design (update the gb file with designed primers)
    3. View the new plasmid map and update lists 
    Run block 1 2 3 subsequently and see what happen after each run
    """
    #
    # { .................... block 1. Edit the sequence (create a new gb file with the BsaI site replaced) -> 'pUCv2'
    filename0 = 'pUC19'
    filename1 = 'pUCv2(1)'
    gb_str = GBparser.Read_gb(filename0)
    GBparser.check_against_IIS_site(gb_str, IIS_site='GGTCTC', display=False)   # show that a 'GGTCTC' is inreversely positioned at range(1765, 1771)
    gb_str = GBparser.gb_edit(gb_str,
                              j0=1765,
                              j1=1771,  # replace 'GGTCTC' reversely positioned at range(1765, 1771)
                              ins_seq='GCTCTC',  # with 'GcTCTC'
                              ins_rf='r',  # also in reverse (r)
                              ins_type='misc_feature',
                              ins_annotation='remove IIS site (silent mutation in AmpR)')
    gb_str = GBparser.gb_reindex(gb_str, origin='tttacactttatgcttccggctcgtatgttg')  # reindex (and reverse) by lac promoter
    gb_str = GBparser.output_gb(gb_str, new_locus_name=filename1, new_definition=filename1)
    GBparser.gb_display(gb_str)    # show that the edited region is now forwardly positioned at range(1458,1464)
    # }

    # { .................... block 2. Cloning design (update the gb file with designed primers)
    filename1 = 'pUCv2(1)'
    GBparser.cloning_design_wrapper_0(
        initial_filename=filename1,
        new_filename=filename1,     # overwrite
        mutation_site=[1459],
        pcr_source=['pUC19'],
        oligo_prefix='Oz',
        N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0,
        verbose=False)
    # }

    # { .................... block 3. check the resulting file & update list
    filename1 = 'pUCv2(1)'
    GBparser.gb_display(GBparser.Read_gb(filename1), display_range = (1359,1559))
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }





    # ................................................................................ Example 0b: subclone a gene onto a vector
    """
    1. Edit the sequence (create a gb file with a the lacZ-alpha replaced with cat gene) -> pUCv2-cat
    2. Cloning design (update the gb file with cloning primers flanking the cat gene)
    3. View the new plasmid map and update lists 
    Run block 1 2 3 subsequently and see what happen after each run
    """
    # { .................... block 1. Edit the sequence (create a gb file with a the lacZ-alpha replaced with cat gene) -> pUCv2-cat
    filename0 = 'pUCv2'
    filename1 = 'pUCv2-cat(1)'
    cat_gene = '''atggagaaaaaaatcactggatataccaccgttgatatatcccaatggcatcgtaaagaacattttgaggcatttcagtcagttgctcaatgtacctataaccagaccgttcagctggatattacggcctttttaaagaccgtaaagaaaaataagcacaagttttatccggcctttattcacattcttgcccgcctgatgaatgctcatccggaattccgtatggcaatgaaagacggtgagctggtgatatgggatagtgttcacccttgttacaccgttttccatgagcaaactgaaacgttttcatcgctctggagtgaataccacgacgatttccggcagtttctacacatatattcgcaagatgtggcgtgttacggtgaaaacctggcctatttccctaaagggtttattgagaatatgtttttcgtctcagccaatccctgggtgagtttcaccagttttgatttaaacgtggccaatatggacaacttcttcgcccccgttttcaccatgggcaaatattatacgcaaggcgacaaggtgctgatgccgctggcgattcaggttcatcatgccgtctgtgatggcttccatgtcggcagaatgcttaatgaattacaacagtactgcgatgagtggcagggcggggcgtaa'''  # from pACYC184 vector

    gb_str = GBparser.Read_gb(filename0)
    gb_str = GBparser.gb_edit(gb_str,
                              j0=74,
                              j1=398,  # replace lacZ-alpha forwardly positioned at range(74,398)
                              ins_seq=cat_gene,  # with cat_gene
                              ins_rf='f',  # also in forward (f)
                              ins_type='CDS',
                              ins_annotation='cat')
    gb_str = GBparser.output_gb(gb_str, new_locus_name=filename1, new_definition=filename1)
    GBparser.gb_display(gb_str)  # show that the cat gene is now forwardly positioned at range(74,734)
    # }

    # { .................... block 2. Cloning design (update the gb file with cloning primers flanking the cat gene)
    filename1 = 'pUCv2-cat(1)'
    GBparser.cloning_design_wrapper_0(
        initial_filename=filename1,
        new_filename=filename1,     # overwrite
        mutation_site=[74, 734],     # list the insert-to-vector connection sites (integers)
        pcr_source=['pUCv2', 'pACYC184'],
        oligo_prefix='Oz',
        IIS_side=['l3', 'r3'],   # so that vector-specific oligos can be reused for subcloning other genes
        N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0,
        verbose=False)
    # }

    # { .................... block 3. View the new plasmid map and update lists
    filename1 = 'pUCv2-cat(1)'
    GBparser.gb_display(GBparser.Read_gb(filename1))
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }





    # ................................................................................ Example 0c: construct a fusion protein involving primer walk
    """
    1. Edit the sequence by index (clone cat gene from pACYC vector and attach an N-terminal SUMO gene by synthetic oligo)
    2. Cloning design by index (update the gb file with cloning primers)
    3. View the new plasmid map and update lists 
    Run block 1 2 3 subsequently and see what happen after each run
    """
    # { .................... block 1. Edit the sequence (clone cat gene from pACYC vector and attach an N-terminal SUMO gene by synthetic oligo)
    filename0 = 'pUCv2'
    filename1 = 'pUCv2-SUMO-cat(1)'
    cat_gene = '''atggagaaaaaaatcactggatataccaccgttgatatatcccaatggcatcgtaaagaacattttgaggcatttcagtcagttgctcaatgtacctataaccagaccgttcagctggatattacggcctttttaaagaccgtaaagaaaaataagcacaagttttatccggcctttattcacattcttgcccgcctgatgaatgctcatccggaattccgtatggcaatgaaagacggtgagctggtgatatgggatagtgttcacccttgttacaccgttttccatgagcaaactgaaacgttttcatcgctctggagtgaataccacgacgatttccggcagtttctacacatatattcgcaagatgtggcgtgttacggtgaaaacctggcctatttccctaaagggtttattgagaatatgtttttcgtctcagccaatccctgggtgagtttcaccagttttgatttaaacgtggccaatatggacaacttcttcgcccccgttttcaccatgggcaaatattatacgcaaggcgacaaggtgctgatgccgctggcgattcaggttcatcatgccgtctgtgatggcttccatgtcggcagaatgcttaatgaattacaacagtactgcgatgagtggcagggcggggcgtaa'''  # from pACYC184 vector
    SUMO_pp = 'MSDSEVNQEAKPEVKPEVKPETHINLKVSDGSSEIFFKIKKTTPLRRLMEAFAKRQGKEMDSLRFLYDGIRIQADQTPEDLDMEDNDIIEAHREQIGGATY'
    SUMO_gene = GBparser.codon_optimization(pp=SUMO_pp, species='ecoli')

    gb_str = GBparser.Read_gb(filename0)
    gb_str = GBparser.gb_edit(gb_str,
                              j0=74,
                              j1=398,  # replace lacZ-alpha forwardly positioned at range(74,398)
                              ins_seq=cat_gene,  # with cat_gene
                              ins_rf='f',  # in forward (f)
                              ins_type='CDS',
                              ins_annotation='cat')
    gb_str = GBparser.gb_edit(gb_str,
                              j0=74,
                              j1=74,  # insert before the cat gene
                              ins_seq=SUMO_gene,  # with a SUMO_gene
                              ins_rf='f',  # in forward (f)
                              ins_type='CDS',
                              ins_annotation='SUMO')
    gb_str = GBparser.output_gb(gb_str, new_locus_name=filename1, new_definition=filename1)
    GBparser.gb_display(gb_str)  # show that the SUMO gene is at range(74,377), the cat gene is at range(377,1037)
    # }

    # { .................... block 2. Cloning design (update the gb file with cloning primers)
    # Since the SUMO_gene is to be assembled from primers, the whole range(2217,2520) shall be listed in addition to insert-to-vector connection sites (int)
    filename1 = 'pUCv2-SUMO-cat(1)'
    GBparser.cloning_design_wrapper_0(
        initial_filename=filename1,
        new_filename=filename1,  # overwrite
        mutation_site=[range(74, 377), 1037],
        pcr_source=['pUCv2', 'pACYC184'],
        oligo_prefix='Oz',
        IIS_side=['l4', 'r4'],
        N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0,
        verbose=False)
    # }

    # { .................... block 3. View the new plasmid map and update lists
    filename1 = 'pUCv2-SUMO-cat(1)'
    GBparser.gb_display(GBparser.Read_gb(filename1))
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }






    # ................................................................................ Example 0d:  subclone a gene onto a vector & mutate BsaI site simultaneously
    """
    1. Edit the sequence by index (clone MBP gene from E. coli genome, mutate the BsaI site on pUC19 in the MBP gene)
    2. Cloning design by index (update the gb file with cloning primers)
    3. View the new plasmid map and update lists 
    Run block 1 2 3 subsequently and see what happen after each run
    """
    # { .................... block 1. Edit the sequence by index (clone MBP gene from E. coli genome, mutate the BsaI site on pUC19 in the MBP gene)
    filename0 = 'pUC19'
    filename1 = 'pUC-MBP(1)'
    mbp_gene = '''atgaaaatcgaagaaggtaaactggtaatctggattaacggcgataaaggctataacggtctcgctgaagtcggtaagaaattcgagaaagataccggaattaaagtcaccgttgagcatccggataaactggaagagaaattcccacaggttgcggcaactggcgatggccctgacattatcttctgggcacacgaccgctttggtggctacgctcaatctggcctgttggctgaaatcaccccggacaaagcgttccaggacaagctgtatccgtttacctgggatgccgtacgttacaacggcaagctgattgcttacccgatcgctgttgaagcgttatcgctgatttataacaaagatctgctgccgaacccgccaaaaacctgggaagagatcccggcgctggataaagaactgaaagcgaaaggtaagagcgcgctgatgttcaacctgcaagaaccgtacttcacctggccgctgattgctgctgacgggggttatgcgttcaagtatgaaaacggcaagtacgacattaaagacgtgggcgtggataacgctggcgcgaaagcgggtctgaccttcctggttgacctgattaaaaacaaacacatgaatgcagacaccgattactccatcgcagaagctgcctttaataaaggcgaaacagcgatgaccatcaacggcccgtgggcatggtccaacatcgacaccagcaaagtgaattatggtgtaacggtactgccgaccttcaagggtcaaccatccaaaccgttcgttggcgtgctgagcgcaggtattaacgccgccagtccgaacaaagagctggcaaaagagttcctcgaaaactatctgctgactgatgaaggtctggaagcggttaataaagacaaaccgctgggtgccgtagcgctgaagtcttacgaggaagagttggtgaaagatccgcgtattgccgccactatggaaaacgcccagaaaggtgaaatcatgccgaacatcccgcagatgtccgctttctggtatgccgtgcgtactgcggtgatcaacgccgccagcggtcgtcagactgtcgatgaagccctgaaagacgcgcagacttga'''  # from E. coli genome

    gb_str = GBparser.Read_gb(filename0)
    gb_str = GBparser.gb_reindex(gb_str, origin='tttacactttatgcttccggctcgtatgttg')  # reindex (and reverse) by lac promoter
    gb_str = GBparser.gb_edit(gb_str,
                              j0=74,
                              j1=398,  # replace lacZ-alpha forwardly positioned at range(74,398)
                              ins_seq=mbp_gene,  # with cat_gene
                              ins_rf='f',  # in forward (f)
                              ins_type='CDS',
                              ins_annotation='MBP')

    GBparser.check_against_IIS_site(gb_str, IIS_site='GGTCTC', display=True)   # show that a 'GGTCTC' is forwardly positioned at range(131, 137) & (2238, 2244)

    gb_str = GBparser.gb_edit(gb_str,
                              j0=133,
                              j1=134,  # replace the 3rd nucleotide of the first BsaI site
                              ins_seq='g',  # with g (silent mutation in MBP gene)
                              ins_rf='f',  # in forward (f)
                              ins_type='variation',
                              ins_annotation='remove_BsaI_site')

    gb_str = GBparser.gb_edit(gb_str,
                              j0=2239,
                              j1=2240,  # replace the 2nd nucleotide of the second BsaI site
                              ins_seq='c',  # with c (silent mutation in AmpR)
                              ins_rf='f',  # in forward (f)
                              ins_type='variation',
                              ins_annotation='remove_BsaI_site')

    gb_str = GBparser.output_gb(gb_str, new_locus_name=filename1, new_definition=filename1)
    GBparser.gb_display(gb_str)     # show that the MBP gene is now forwardly positioned at range(74,1178)
    # }

    # { .................... block 2. Cloning design (update the gb file with cloning primers)
    filename1 = 'pUC-MBP(1)'
    GBparser.cloning_design_wrapper_0(
        initial_filename=filename1,
        new_filename=filename1,  # overwrite
        mutation_site=[74, 133, 1178, 2239],
        pcr_source=['pUC19', 'ecoli', 'ecoli', 'pUC19'],
        oligo_prefix='Oz',
        IIS_side=['l4', 'u', 'r4', 'u'],
        N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0,
        verbose=False)
    # }

    # { .................... block 3. View the new plasmid map and update lists
    filename1 = 'pUC-MBP(1)'
    GBparser.gb_display(GBparser.Read_gb(filename1))
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }




import os
from biochemHH import GBparser
from biochemHH.GBparser import reverse_complement
from biochemHH.GBparser import preferred_codon
from biochemHH.GBparser import preferred_codon_split
from biochemHH.GBparser import Read_gb

if __name__ == '__main__':


    wd = 'C:/Users/hsinmei/playground'; os.chdir(wd)

    """ 
    This example file consist of the following parts:
    
    1. Primer design
        - Analyze a forward primer OR a primer pair
        - Design primer pairs
        - Analyze an oligo's secondary structure propensity
        - Pick overlapping regions for primer walk
        - Manually pick the annealing regions for golden gate cloning
        
    2. Codon choice 
        (supported species: 'ecoli', 'ecoliK', 'bacillus', 'yeast', 'pichia', 'mouse', 'hamster', 'human')
        - Codon optimization
        - Codon diagnosis
        - Choose a degenerate codon
        - Create a dictionary of preferred codon
        
    3. GB editing tools (BATCH)
        - Remove primer annotations from all GB files
        - Edit & annotate all GB files containing the target_seq
        - Edit & annotate all GB files containing the target_seq in certain context
        - Simple string replacement applied to all GB files
        
    4. GB editing tools (SINGLE)
        - Duplicate a GB file given a new filename 
        - Reverse the sequence
        - Reindex the sequence
        - Remove features
        - Add features
        - Display by range
        - Display by feature
        - Sequence editing by index
        - Sequence editing by variation of an annotated CDS/gene
    
    5. SEQ to Notation interconversion
        - SEQ to Notation
        - Notation to SEQ
    
    6. Snippets useful in cloning design
        - Check if the sequence contain BsaI site
        - Update list_0 with list_1
        - Batch-mode definition update
        - Full GB display
    """
    # ================================================================================ 1. Primer design

    # { ........................................ Analyze a forward primer OR a primer pair
    p1 = GBparser.Primer3_analysis_250617(
        left_primer = 'tgatATTcttgaggaacgccaaaaagt',
        right_primer =  'atcacccagcaggctaggaataaaacc', # or right_primer =  None
        thermodynamic=0,
        monovalent_cation=50, divalent_cation=1.5,
        verbose=True, horizontal = False)
    # }


    # { ........................................ Design primer pairs
    """
    Advanced: to design an inverse PCR primer pair, insert a comma at the intended site, e.g., f'{seq1},{seq2}' -> f'{seq2+seq1}' 
    """
    template = 'AAAATCGAAGAAGGTAAACTGGTAATCTGG,ATTAACGGCGATAAAGGCTATAACGGTCTCGCTGAA'
    if ',' in template: template = template.split(',')[1] + template.split(',')[0].replace(' ','')
    primers = GBparser.Primer3_design(template, mode=1, primer_max_size=25, primer_pick_anyway=1)
    # }



    # { ........................................ Analyze an oligo's secondary structure propensity
    seq = 'ACTCAAACCCGTGGGCTAGTTTATC'
    GBparser.analyze_single_seq(seq, verbose=True)
    # }


    # { ........................................ Pick overlapping regions for primer walk
    substring = 'CTAGCCCACGGGTTTGAGTGCTCAAGACATGGC'
    a = GBparser.Show_po_pairs(substring, exact = False, preset = 1)
    print(a)
    # }


    # { ........................................ Manually pick the annealing regions for golden gate cloning
    """ 
    Provide a list of N segments (lst_of_seq) within which the annealing region is intended.
    The function will pick a 4-nt region per segment (N) such that the resulting overhangs (2*N) have minimal tendency to mis-anneal.
    """
    GBparser.choose_golden_gate(lst_of_seq=['AACG', 'tatgtg', 'ctagaga'], N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0)
    # }



    # ================================================================================ 2. Codon choice

    # { ........................................ Codon optimization
    #  supported species include 'ecoli', 'ecoliK', 'bacillus', 'yeast', 'pichia', 'mouse', 'hamster', 'human'
    pp = "KIEEGKLVIWINGDKGYNGLAE"
    dna1 = GBparser.codon_optimization(pp = pp , species = 'ecoli',
                                       preceding_triplet='ATG', tailing_triplet='GTC', verbose = True)
    # }


    # { ........................................ Codon diagnosis
    #  supported species include 'ecoli', 'ecoliK', 'bacillus', 'yeast', 'pichia', 'mouse', 'hamster', 'human'
    dna1 = '''TTTCAAATTGTGAAGAATCCAAGGTCTGTGGGAAA
    AGCAAGCGAGCAGCTGGCTGGCAAGGTGGCACAAGTCAAGAAGAACGGAAGAATCAGCCTGGTGCTGGGC
    GGAGACCACAGTTTGGCAATTGGAAGCATCTCTGGCCATGCCAGGGTCCACCCTGATCTTGGAGTCATCT
    GGGTGGATGCTCACACTGATATCAACACTCCACTGACAACCACAAGTGGAAACTTGCATGGACAACCTGT
    '''
    GBparser.codon_diagnosis(dna1, rv_cutoff=0.1, species='ecoli')
    # }


    # { ........................................ Choose a degenerate codon
    # Intended for site-directed semi-random mutagenesis
    target = 'FIL'
    codon_table, _  = GBparser.extract_kazusa()
    df1, codon_detail, codon = GBparser.Degenerate_Chooser(codon_table, target)
    # }


    # { ........................................ Create a dictionary of preferred codon
    #  supported species include 'ecoli', 'ecoliK', 'bacillus', 'yeast', 'pichia', 'mouse', 'hamster', 'human'
    dict1 = GBparser.preferred_triplet(species = 'ecoli')
    for key, value in dict1.items():
        print(key, value)
    # }


    # # ================================================================================ 3. GB editing tools (BATCH)
    """ 
    Will overwrite all matching GB files within the working directory
    """

    # { ........................................ Remove primer annotations from all GB files
    GBparser.remove_primer_feature_for_gb_in_a_folder()
    # }


    # { ........................................ Edit & annotate all GB files containing the target_seq
    GBparser.add_feature_for_gb_in_a_folder(target_seq = 'gggaaacgcctggtatcttt',
                                            type = 'primer_bind',
                                            annotation = 'pBR322ori-F')
    # }


    # { ........................................ Edit & annotate all GB files containing the target_seq in certain context
    #  If there are two '|' characters in target_seq, it will be interpreted as "preceding_seq|target_seq|tailing_seq". Only the target_seq will be replaced
    GBparser.edit_sequence_for_gb_in_a_folder( target_seq = 'aaaaggaagagtatg|agt|attcaa',
                                               new_seq = 'gtg',
                                               ins_type='variation',
                                               ins_annotation='S2V')
    # }


    # { ........................................ Simple string replacement applied to all GB files
    GBparser.string_replace_for_gb_in_a_folder( target_str = 'M13 fwd', replacement_str = 'M13F')
    # }


    # # ================================================================================ 4. GB editing tools (SINGLE)

    # { ......................................... Duplicate a GB file given a new filename
    filename = 'pUC19'
    new_filename = 'pUC19_new'
    gb_str = GBparser.Read_gb(filename)
    gb_str=  GBparser.gb_rename_locus(gb_str, new_filename)
    GBparser.output_gb(gb_str, new_filename)
    # }


    # { ........................................ Reverse the sequence
    filename = 'pUC19'
    new_filename = 'pUC19_new'
    gb_str = GBparser.Read_gb(filename)
    gb_str = GBparser.gb_reverse(gb_str)
    GBparser.output_gb(gb_str, new_filename)
    # }


    # { ........................................ Reindex the sequence
    # By setting promter as index 0, the gene of interest will be forwardly positioned near the start
    filename = 'pUC19'
    new_filename = 'pUC19_new'
    gb_str =  GBparser.Read_gb(filename)
    gb_str =  GBparser.gb_reindex(gb_str, origin='tttacactttatgcttccggctcgtatgttg') # lac promoter
    GBparser.output_gb(gb_str, new_filename)
    # }


    # { ........................................ Remove features
    filename = 'pUC19'
    new_filename = 'pUC19_new'
    gb_str = GBparser.Read_gb(filename)
    gb_str = GBparser.filter_feature(gb_str, remove='source', by= 'type')
    gb_str = GBparser.filter_feature(gb_str, remove= 'MCS', by='label')
    GBparser.output_gb(gb_str, new_filename)
    # }


    # { ........................................ Add features
    """ 
    Sequence within range(s, e) will be annotated. Index start from 0.
    (i.e., to annotate a region, select the region in Benchling or SnapGene, s = start-1, e = end)
    """
    filename = 'pUC19'
    new_filename = 'pUC19_new'
    gb_str = GBparser.Read_gb(filename)
    gb_str = GBparser.gb_add_feature(gb_str, s = 10, e = 100, rf = 'f', type = 'misc_structure', annotation = 'test')
    GBparser.output_gb(gb_str, new_filename)
    # }


    # { ........................................ Display by range
    filename = 'pUC19'
    gb_str = GBparser.Read_gb(filename)
    GBparser.gb_display(gb_str,
                        display_range=(1625, 2486),
                        type_filter=['CDS', 'variation', 'primer'],  # will only show the listed features
                        )
    # }


    # { ........................................ Display by feature
    filename = 'pUC19'
    gb_str = GBparser.Read_gb(filename)
    GBparser.gb_display(gb_str,
                        feature_select='AmpR',  # feature_select will overwrite display_range
                        type_filter=['CDS', 'variation', 'primer'],  # will only show the listed features
                        )
    # }


    # { ........................................ Sequence editing by index
    """
    To replace a region, select the region in Benchling or SnapGene, j0 = start-1, j1 = end.
    To insert, select the nucleotide after the insertion site, j0 = start-1, j1 = start-1 
    """
    filename = 'pUC19'
    new_filename = 'pUC19_new'
    gb_str = GBparser.Read_gb(filename)
    gb_str = GBparser.gb_edit(gb_str,
                              j0=2432,
                              j1=2435,
                              ins_rf='r',
                              ins_seq='GAA',
                              ins_type='misc_feature', ins_annotation='AmpR_C18E')
    GBparser.output_gb(gb_str, new_filename)
    # }



    # { ........................................ Sequence editing by variation of an annotated CDS/gene
    """ 
    Will automatically reverse the sequence if feature_select is on the reverse strand.
    Supported species include 'ecoli', 'bacillus', 'yeast', 'pichia', 'mouse', 'hamster', 'human'
    """
    filename = 'pUC19'
    new_filename = 'pUC19_new'
    gb_str = GBparser.Read_gb(filename)
    Gb_str, *_ = GBparser.gb_edit_variation(gb_str,
                                            feature_select='AmpR',
                                            mutation_codon_no=[24],
                                            replacement_codon=[preferred_codon('A', species='ecoli')],
                                            )
    GBparser.output_gb(Gb_str[0], new_filename)
    # }


    # ================================================================================ 5. SEQ to Notation interconversion
    """
       Variation consist of 6 types 
           1-aa substitution, e.g.,    R13A or 13A  
           1-aa deletion, e.g.,        28del
           1-aa insertion, e.g.,       90~91insE
           multi insertions, e.g.,     101~102insEAT
           multi deletion, e.g.,       130~135del
           deletion & insertion, e.g., 140~142delinsAAG
           
       Multiple variations were joined by '_' to form a Notation
    """

    # { ........................................ SEQ to Notation
    # Compare the difference between ref & seq, and output a notation
    ref = 'MEKKITGYTTVDISQWHRKEHFEAFQSVAQCTYNQTVQLDITAFLKTVKKNKHKFYPAFIHILARLMNAHPEFRMAMKDGELVIWDSVHPCYTVFHEQTETFSSLWSEYHDDFRQFLHIYSQDVACYGENLAYFPKGFIENMFFVSANPWVSFTSFDLNVANMDNFFAPVFTMGKYYTQGDKVLMPLAIQVHHAVCDGFHVGRMLNELQQYCDEWQGGA'
    seq = 'MEKKITGYTTVDASQWHRKEHFEAFQSVAQCYNQTVQLDITAFLKTVKKANKHKFYPAFIHILARLPEFRMAMKDGELVIAGARWDSVHPCYTVFHFINELLAFSSLWSEYHDDFRQFLHIYSQDVACYGENLAYFPKGFIENMFFVSANPWVSFTSFDLNVANMDNFFAPVFTMGKYYTQGDKVLMPLAIQVHHAVCDGFHVGRMLNELQQYCDEWQGGA'

    notation = GBparser.seq_to_notation(ref, seq, force_substitution=False)
    print(notation)
    # }


    # { ........................................  Notation to SEQ
    # create a variant (seq) based on a ref and a notation
    ref = 'MEKKITGYTTVDISQWHRKEHFEAFQSVAQCTYNQTVQLDITAFLKTVKKNKHKFYPAFIHILARLMNAHPEFRMAMKDGELVIWDSVHPCYTVFHEQTETFSSLWSEYHDDFRQFLHIYSQDVACYGENLAYFPKGFIENMFFVSANPWVSFTSFDLNVANMDNFFAPVFTMGKYYTQGDKVLMPLAIQVHHAVCDGFHVGRMLNELQQYCDEWQGGA'
    notation1 = '_I13A_32del_50~51insA_67~70del_84~85insAGAR_97~101delinsFINELLA'

    seq1 = GBparser.notation_to_seq(ref, notation1)
    print(seq1)
    # }



    # { ........................................  Notation to SEQ to Notation
    # create a variant (seq) based on a ref and a notation
    ref = 'MEKKITGYTTVDISQWHRKEHFEAFQSVAQCTYNQTVQLDITAFLKTVKKNKHKFYPAFIHILARLMNAHPEFRMAMKDGELVIWDSVHPCYTVFHEQTETFSSLWSEYHDDFRQFLHIYSQDVACYGENLAYFPKGFIENMFFVSANPWVSFTSFDLNVANMDNFFAPVFTMGKYYTQGDKVLMPLAIQVHHAVCDGFHVGRMLNELQQYCDEWQGGA'
    notation = '_13A_52T'

    seq1 = GBparser.notation_to_seq(ref, notation)
    print(seq1)

    notation1 = GBparser.seq_to_notation(ref, seq1, force_substitution=True)
    print(notation1)

    # }
    # ================================================================================ 6. Snippets useful in cloning design

    # { ........................................ Check if the sequence contain BsaI site
    GBparser.check_against_IIS_site(gb_str=GBparser.Read_gb('pUC19'), IIS_site='GGTCTC', display=True)
    # }


    # { ........................................ Update list_0 with list_1
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }


    # { ........................................ Batch-mode definition update
    # Put the name_definition.csv in the working directory
    GBparser.edit_definition_for_gb_in_a_folder(guiding_file='name_definition.csv')
    # }


    # { ........................................ Full GB display
    GBparser.gb_display(GBparser.Read_gb('pUC19'))
    # }


import os
from time import perf_counter
from biochemHH import GBparser
from biochemHH.GBparser import reverse_complement
from biochemHH.GBparser import preferred_codon
from biochemHH.GBparser import preferred_codon_split

if __name__ == '__main__':
    wd = 'C:/Users/hsinmei/playground';
    os.chdir(wd)

    """
    This example file consist of the following parts:
    1. Print notes
    2. Cloning design (batch-mode)
        - Example 1a: saturated mutagenesis at a single site
        - Example 1b: alanine scanning at multiple sites
        - Example 1c: cloning design based on a list of notations
    """

    # ================================================================================ 1. Print notes (batch-mode)
    # { ........................................ Print notes (simple)
    GBparser.print_note('Cloning design (batch-mode)')
    GBparser.print_note('About oligo & plasmid list')
    GBparser.print_note('About notation')
    # }

    # { ........................................ Print notes (advanced)
    GBparser.print_note('About golden gate')
    GBparser.print_note('Primer design explained')
    # }

    # ================================================================================ 2. Cloning design (batch-mode)
    """ 
    Design multiple gene variants according to the amino acid index (starting at 1).
    
    Required input: a plasmid GB file (base_filename) containing the annotated gene/CDS (feature_select).
    
    Codons are selected automatically based on the expression host. Supported species include: 'ecoli', 'ecoliK', 'bacillus', 'yeast', 'pichia', 'mouse', 'hamster', and 'human'
    """

    # { ........................................ Example 1a: saturated mutagenesis at a single site
    GBparser.cloning_design_wrapper_1(
        base_filename='pUCv2-cat',  # the gb file 'pUCv2-cat.gb' shall be present in the working directory
        feature_select='cat',
        mutation_codon_no=[16],
        replacement_codon=preferred_codon_split('RHKDESTNQAVLIMFYGPCW', species='ecoli'),
        plasmid_prefix='pL-',
        oligo_prefix='Oz',
        N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0,
        verbose=False)
    # }

    # { .................... after check
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }





    # { ........................................ Example 1b: alanine scanning at multiple sites
    GBparser.cloning_design_wrapper_1(
        base_filename='pUCv2-cat',
        feature_select='cat',
        mutation_codon_no=[41, 42, 44, 45],
        replacement_codon=preferred_codon('A', species='ecoli'),  # A for alanine.
        plasmid_prefix='pL-',
        oligo_prefix='Oz',
        N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0,
        verbose=False)
    # }

    # { .................... after check
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }





    # { ........................................ Example 1c: cloning design based on a list of notations
    """
    Variations consist of 6 types
        1-aa substitution, e.g.,    R13A or 13A  
        1-aa deletion, e.g.,        28del
        1-aa insertion, e.g.,       90~91insE
        multi insertions, e.g.,     101~102insEAT
        multi deletion, e.g.,       130~135del
        deletion & insertion, e.g., 140~142delinsAAG
    Multiple variations are joined by '_' to form a notation, e.g., '_13A_28del_90~91insE'
    """

    notations = ['_13A_28del_90~91insE',
                 '_101~102insEAT_130~135del_140~142delinsAAG']

    GBparser.cloning_design_wrapper_1(
        base_filename='pUCv2-cat',
        feature_select='cat',
        notations=notations,
        species='ecoli',
        plasmid_prefix='pL-',
        oligo_prefix='Oz',
        N_match_cutoff=2, YR_match_cutoff=4, GC_threshold=0,
        verbose=False)
    # }

    # { .................... after check
    GBparser.update_list('Oligo')
    GBparser.update_list('Plasmid')
    # }



import os
from biochemHH import Homolog
from biochemHH import StructureHH
from biochemHH import GBparser


if __name__ == '__main__':

    wd = 'C:/Users/hsinmei/playground'; os.chdir(wd)

    """
    This example file consist of the following parts:
    
    1. Edit a single PDB(CIF) file
        - Duplicate a structure
        - Extract certain residue (ligand) from a structure
        - Substitute the nucleobase
        - Rename chains
        - Delete chains
        - Edit residue: re-assign, rename, and keep all atoms
        - Edit residue: remove certain atoms
        - Edit chain: retain a list of residues
        - Edit chain: remove a list of residues
        - Duplicate a residue
        - ATOM to HETATM (some AI tools only consider HETATM as ligand)
      
    2. Align & merge two PDB(CIF) files
        - Chimera example 1, alignment by oligonucleotide
        - Chimera example 2, alignment by amino acid
        - Chimera example 3, alignment by atoms
    """

    # ////////////////////////////////////////////////////////////////////////////////  Edit a single PDB(CIF) file
    # { ........................................ Duplicate a structure
    """ 
    Useful in case of successive editing.
    """
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose =  True
    StructureHH.copy_structure_file(filename0, filename1, verbose = verbose)

    # }


    # { ........................................ Extract certain residue (ligand) from a structure
    """
    The resulting .cif will be exported to the working directory
    """
    filename0 = '2W35.cif'
    chain = 'C'
    resi = 8
    verbose = True
    atoms_to_remove = ["C2'", "C3'", "C4'", "C5'", "O2'", "O3'", "O4'", "O5'",
                       "H1'", "H2'", "H2''", "H3'", "H4'", "H5'", "H5''", "HO2'", "HO3'", "P", "OP1", "OP2"]

    StructureHH.create_residue_cif(filename0 ,
                                   chain = chain,
                                   resi = resi,
                                   atoms_to_remove = atoms_to_remove,
                                   verbose = verbose)
    # }


    # { ........................................ Substitute the nucleobase
    """
    Available change_to include 'adenine', 'thymine',  'cytosine' , 'guanine', 'uracil', 'inosine'
    Below is an example of successive editing.
    """
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.base_substitution(filename1,
                                  chain='A',
                                  resi=304,
                                  change_to='uracil')

    StructureHH.base_substitution(filename1,
                                  chain='P',
                                  resi=72,
                                  change_to='adenine',
                                  verbose=verbose)

    # }


    # { ........................................ Rename chains
    """
    The example rename chain 'A' -> 'C', chain 'B' -> 'D' respectively.
    """
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.rename_chains(filename1,
                              initial_chain_id=['A', 'B'],
                              new_chain_id=['C', 'D'],
                              verbose= verbose)
    # }


    # { ........................................ Delete chains
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.delete_chains(filename1, chains_to_remove = ['B'], verbose = verbose)
    # }


    # { ........................................ Edit residue: re-assign, rename, and keep all atoms
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.edit_residue(filename1,
                             initial_chain_resi = ['A', 304],
                             new_chain_resi = ['L', 1],
                             new_resname = 'LIG',
                             remove_atoms = None,
                             verbose = verbose)
    # }


    # { ........................................ Edit residue: remove certain atoms
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.edit_residue(filename1,
                             initial_chain_resi = ['A', 304],
                             new_chain_resi = None,
                             new_resname = None,
                             remove_atoms = ["O2'", "O3'"],
                             verbose = verbose)
    # }


    # { ........................................ Edit chain: retain a list of residues
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.chain_shrinking(filename1,
                                chain = 'A',
                                resi_to_retain = range(100,200),
                                verbose = verbose)
    # }


    # { ........................................ Edit chain: remove a list of residues
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.chain_shrinking(filename1,
                                chain = 'A',
                                resi_to_remove = range(10,200),
                                verbose = verbose)
    # }


    # { ........................................ Duplicate a residue
    """
    the duplicated residue will be attached to the end of the same chain
    """
    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.duplicate_residue(filename1,
                                  initial_chain_resi = ['A', 304],
                                  verbose=verbose)
    # }


    # { ........................................ ATOM to HETATM (some AI tools only consider HETATM as ligand)

    filename0 = '5AXN.cif'
    filename1 = 'temporary.pdb'
    verbose = True
    StructureHH.copy_structure_file(filename0, filename1)

    StructureHH.atom_to_hetatm(filename1,
                               chain_id = 'P',
                               verbose = verbose)
    # }



    # ////////////////////////////////////////////////////////////////////////////////  Align & merge two PDB(CIF) files

    ''' 
    The function merges structures t0 and t1, aligning t1 to t0 based on specified residues (t0_align and t1_align).
        For polypeptide residues, alignment is performed using atoms ["N", "CA", "C"].
        For DNA or RNA residues, alignment is performed using atoms ["C1'", "C2'", "C3'", "C4'", "O4'"].
        To skip alignment, set t0_align = None.

    You can choose to retain specific residues from t0 (t0_sele) and t1 (t1_sele). For example:
        To retain all residues of all chains in t0, set t0_sele = 'all'.
        To retain all residues in chain A, residues 1–23 in chain B, and residues 2, 4, and 6 in chain P, set t0_sele = "A B1~23 P2 P4 P6".

    Note: After merging, chains from t0 and t1 will be renamed in alphabetical order.
    '''

    # { ........................................ Chimera example 1, alignment by oligonucleotide
    StructureHH.make_chimera(
        t0='5AXN.cif',
        t0_align = 'P2 P4 P6',      # align by reisude 2, 4, 6 in chain P
        t0_sele = 'A B1~23 P2~6 P49~72',    # retain all residues in chain A, residue 1~23 in chain B, resideu 49~72 in c hain P

        t1 = '1FIX.cif',
        t1_align = 'A2 A4 A6',      # align by reisude 2, 4, 6 in chain A
        t1_sele = 'all',    # retain all residues from all chains

        new_filename= 'chimera1.pdb')
    # }



    # { ........................................ Chimera example 2, alignment by amino acid
    StructureHH.make_chimera(
        t0='2PYL.cif',
        t0_align='A8~136 A483~574',
        t0_sele='all',

        t1='2EX3.cif',
        t1_align='A8~136 A483~574',  # be careful all the residues shall be in structured region
        t1_sele='A B',

        new_filename='chimera2.pdb')
    # }



    # { ........................................ Chimera example 3, alignment by atoms
    ''' 
    The function will merge structure t0 & t1, among which t1 will be aligned to t0 by a list of atoms (t0_atoms, t1_atoms).   
    Specify each atom by (chain, resi, atom.id)
    '''
    t0_atoms = [('A', 303, 'MG'),
                ('A', 302, 'MG'),
                ('A', 301, 'MG'),
                ('A', 304, 'PG'),
                ('A', 304, 'PB'),
                ('A', 304, 'PA'),
                ]

    t1_atoms = [('A', 405, 'MG'),
                ('A', 404, 'MG'),
                ('A', 403, 'MG'),
                ('A', 402, 'PG'),
                ('A', 402, 'PB'),
                ('A', 402, 'PA'),
                ]

    StructureHH.make_chimera_by_atom_aln(
        t0='5AXN.cif',
        t0_atoms=t0_atoms,
        t0_sele='A B P2~7 P49~72',

        t1='3WC0.cif',
        t1_sele='A',
        t1_atoms=t1_atoms,

        new_filename='chimera3.pdb')
    # }






from setuptools import setup, find_packages
from pathlib import Path

this_dir = Path(__file__).parent
long_description = (this_dir / "README.md").read_text(encoding="utf-8")

setup(
    name="biochemHH",
    version="0.0.2",
    author="Otter Brown",
    author_email="",  # hidden for now
    description="Biochemistry & molecular biology helper tools",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    package_data={
        "biochemHH": ["example_input/*", "example_script/*"],
    },
    include_package_data=True,
    install_requires=[
    "biopython",
    "numpy",
    "scipy",
    "matplotlib",
    "pandas",
    "primer3-py",
    "pillow",
    "chardet",
    ],
    python_requires=">=3.8",
    license="GPL-2.0-only",
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    keywords="biochemistry molecular biology bioinformatics",
)
